// dialog_ctor.C

/******************************************************************************
 *
 *  MiXViews - an X window system based sound & data editor/processor
 *
 *  Copyright (c) 1993, 1994 Regents of the University of California
 *
 *  Author:     Douglas Scott
 *  Date:       December 13, 1994
 *
 *  Permission to use, copy and modify this software and its documentation
 *  for research and/or educational purposes and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation. The author reserves the right to distribute this
 *  software and its documentation.  The University of California and the author
 *  make no representations about the suitability of this software for any 
 *  purpose, and in no event shall University of California be liable for any
 *  damage, loss of data, or profits resulting from its use.
 *  It is provided "as is" without express or implied warranty.
 *
 ******************************************************************************/


#ifdef __GNUG__
#pragma implementation
#endif

#include "localdefs.h"
#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/button.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/painter.h>
#include "application.h"
#include "buttonarray.h"
#include "dialog_ctor.h"
#include "eventforwarder.h"
#include "queryvalue.h"
#include "querylist.h"
#include "request.h"
#include "dialogbox.h"
#include "fileselector.h"
#include "textinput.h"
#include "valueslider.h"
#include "vmessage.h"
#include "contscroller.h"

// ResponseButton is a private subclass of PushButton whose label is
// highlighted when the particular instance is the default response to a 
// query, i.e., the one that will be activated if the user presses <return>

class ResponseButton : public PushButton {
public:
	ResponseButton(const char* text, ButtonState* s, int v,
			boolean isDefault=false)
		: PushButton(text, s, v), is_default(isDefault) {}
	ResponseButton(const char* text, ButtonState* s, void* v,
			boolean isDefault=false)
		: PushButton(text, s, v), is_default(isDefault) {}
	virtual ~ResponseButton() {}
	redefined void Refresh();
protected:
	boolean is_default;	// is this button the default for <return>
};

void
ResponseButton::Refresh() {
	int oldStyle = output->GetStyle();
	// button that is pressed when <return> is hit has bold font
	if(is_default) output->SetStyle(Boldface);
	PushButton::Refresh();
	if(is_default) output->SetStyle(oldStyle);
}

//********

// EnableList is a list of Enableable objects bound to a value tag.  Objects whose tag matches
// the passed-in value are enabled;  all others are disabled.

class EnableList {
public:
	EnableList();
	~EnableList();
	void Insert(Enableable *e, int tag);
	void Notify(int newState);
private:
	struct Elem {
		Elem(Enableable *item, int value) : _item(item), _value(value), _next(nil) {}
		Enableable *_item;
		int _value;
		struct Elem *_next;
	} *_list;
};

EnableList::EnableList() : _list(nil) {}

EnableList::~EnableList() {
	Elem *e = _list;
	while (e != nil) {
		Elem *next = e->_next;
		delete e;
		e = next;
	}
}

void
EnableList::Insert(Enableable *e, int tag) {
	Elem *elem = new Elem(e, tag);
	elem->_next = _list;
	_list = elem;
}

void
EnableList::Notify(int newState) {
	Elem *e = _list;
	while (e != nil) {
		e->_item->enable(e->_value == newState);
		e = e->_next;
	}
}


//********

// EntryBox is a private subclass of MonoScene which responds to the changes in state of
// its component radio buttons.

class EntryBox : public MonoScene {
public:
	EntryBox(ButtonState *, QueryExclusiveValues *, EnableList *);
	~EntryBox();
	redefined void Update();
	redefined Interactor *Wrap(Interactor *);
protected:
    ButtonState *			_subject;
private:
	QueryExclusiveValues *	_values;
	EnableList *			_list;
};

EntryBox::EntryBox(ButtonState* state, QueryExclusiveValues* values, EnableList *list)
	: _subject(state), _values(values), _list(list)
{
	_subject->Attach(this);
}


EntryBox::~EntryBox() {
	_subject->Detach(this);
	delete _list;
}

void
EntryBox::Update() {
	int stateValue;
	_subject->GetValue(stateValue);
	_values->setSelection(stateValue);
	_list->Notify(stateValue);
}

Interactor *
EntryBox::Wrap(Interactor *i) {
	return new MarginFrame(i,
					10, 0, 0,			// horiz border 10, fixed
					0, 0, vfil);			// vert  border 0, stretchable
}

//********

DialogConstructor* DialogConstructor::ctorInstance = nil;

DialogConstructor*
DialogConstructor::getInstance() {
	if (ctorInstance == nil)
		ctorInstance = new DialogConstructor;
	return ctorInstance;
}

void
DialogConstructor::destroyInstance() {
	delete ctorInstance;
	ctorInstance = nil;
}

DialogBox *
DialogConstructor::createDialog(Interactor* under, Request &request) {
	typedef boolean (DialogBox::*(DialogForwardFunction))(Event&);
	class DialogEventForwarder : public EventForwarder {
	public:
		DialogEventForwarder(DialogBox* db, DialogForwardFunction dff)
			: myDialog(db), myFunction(dff) {}
		redefined boolean handle(Event &e) {
			return (myDialog->*myFunction)(e);
		}
	private:
		DialogBox* myDialog;
		DialogForwardFunction myFunction;
	};
	typedef DialogBox*
		(*DialogFactoryFunction)(ButtonState*, Interactor*, Response);
	static DialogFactoryFunction dialogFactory[] = {
		&Alert::create,
		&Confirmer::create,
		&ChoiceDialog::create,
		&InputDialog::create
	};
	// all confirm, deny buttons are tied to this state
	ButtonState* state = new ButtonState(NullResponse);
	Response response = request.buttonList->defaultResponse();
	DialogBox* dialog= (*dialogFactory[request.type()])(state, under, response);
	VBox* interior = new VBox;
	if(request.hasLabels()) {
		interior->Insert(messageBox(request.labelList));
	}
	if(request.hasBrowser()) {
		interior->Insert(new HBorder);
		Scene* selector = fileSelector(request.browserQuery, state);
		interior->Insert(selector);
		FileSelector* fs = (FileSelector *) selector;
		fs->setForwarder(
			new DialogEventForwarder(dialog, &DialogBox::eventLoop)
		);
		((InputDialog *) dialog)->addFileSelector(fs);
	}
	if(request.hasValues()) {
		interior->Insert(new HBorder);
		interior->Insert(textInputBox((InputDialog *) dialog,
			request.valueList, state));
	}
	if(request.hasChoices()) {
		interior->Insert(new HBorder);
		interior->Insert(choiceButtonBox(request.choiceList, state));
	}
	interior->Insert(new HBorder);
	// all dialogs have a button box
	interior->Insert(new VGlue(5, 0));	// spacer above buttons
	interior->Insert(buttonBox(request.buttonList, state));
	interior->Insert(new VGlue(2, 0));	// spacer below buttons
	dialog->Insert(interior);
	dialog->setBell(request.useBell());
	Resource::unref(state);
	return dialog;
}

Scene *
DialogConstructor::messageBox(QueryList *list) {
	VBox* mesgBox = new VBox;
	mesgBox->Align(Center);
	for(list->start(); list->more(); list->next()) {
		mesgBox->Insert(
			new MarginFrame(
				new Message("Title", list->item()->label()),
				5, 0, hfil,
				0, 0, vfil
			)
		);
	}
	return mesgBox;
}

Scene *
DialogConstructor::fileSelector(QueryFile *qfile, ButtonState* response) {
	return new FileSelector(
		response, qfile->path(), qfile->suffixes(), 15, 32
	);
}

Scene *
DialogConstructor::textInputBox(InputDialog *ipd, QueryList *list, ButtonState *bs)
{
	VBox* textBox = new VBox;
	const char* inputSubtitle = list->label();
	// insert text entry subtitle before inserting text entry items
	if(inputSubtitle != nil && strlen(inputSubtitle) > 0) {
		textBox->Insert(
			new MarginFrame(
				new Message("Subtitle", inputSubtitle),
				10, 0, 0,		// horiz border 10, fixed
				0, 0, vfil		// vert  border 0, stretchable
			)
		);
	}
	for(list->start(); list->more(); list->next()) {
		QueryItem *item = list->item();
		TextInput* txt = nil;
		ValueSlider* slider = nil;
		
		switch (item->type()) {
		case QueryValue::RangedValue:
			{
				register QueryValue *qvalue = (QueryValue *) item;
				textBox->Insert(
					new MarginFrame(
						slider = new ValueSlider(bs, qvalue),
						10, 0, 0, 0, 0, vfil
					)
				);
				txt = slider->getTextInput();
				// add current TextInput to the linked list of TextInputs
				ipd->addTextInput(txt);
			}
			break;
		case QueryValue::Value:
			{	
				register QueryValue *qvalue = (QueryValue *) item;
				const char* samp = "MMMMMMMMMMMM";
				txt = new TextInput(bs, samp, qvalue);
				textBox->Insert(
					new MarginFrame(
						new HBox(
							new Message("TextEntryLabel", qvalue->label(), Center),
							new HGlue(5, 0),
							new Frame(txt, 1)
						),
						10, 0, 0,			// horiz border 10, fixed
						0, 0, vfil			// vert  border 0, stretchable
					)
				);
				// add current TextInput to the linked list of TextInputs
				ipd->addTextInput(txt);
			}
			break;
		case QueryValue::ExclusiveValues:
			{	
				register QueryExclusiveValues *qexvalues = (QueryExclusiveValues *) item;
				ButtonState* qstate = new ButtonState(qexvalues->defaultSelection());
				QueryList *exlist = qexvalues->getEntryList();
				ChoiceValue buttonValue = 0x1;	// Value for first button
				EnableList *enableList = new EnableList;
				EntryBox *entryBox = new EntryBox(qstate, qexvalues, enableList);
				VBox *exBox = new VBox;
				const char* exLable = item->label();
				if(exLable != nil && strlen(exLable) > 0) {
					exBox->Insert(
						new MarginFrame(
							new Message("ExclusiveValueLabel", exLable),
							10, 0, 0,		// horiz border 10, fixed
							0, 0, vfil		// vert  border 0, stretchable
						)
					);
				}
				for (exlist->start(); exlist->more(); exlist->next()) {
					QueryValue *qvalue = (QueryValue *) exlist->item();
					RadioButton *selectButton = new RadioButton("", qstate, buttonValue);
					bool enabled = (buttonValue == qexvalues->defaultSelection());
					exBox->Insert(new VGlue(2, 0));
					switch (qvalue->type()) {
					case QueryValue::RangedValue:
						exBox->Insert(
							new HBox(
								selectButton,
								new HGlue,
								slider = new ValueSlider(bs, qvalue)
							)
						);
						enableList->Insert(slider, buttonValue);
						slider->enable(enabled);
						// add current TextInput to the linked list of TextInputs
						ipd->addTextInput(slider->getTextInput());
						break;
					case QueryValue::Value:
						{	
							const char* samp = "MMMMMMMMMMMM";
							VMessage *textLabel;
							txt = new TextInput(bs, samp, qvalue);
							exBox->Insert(
								new HBox(
									selectButton,
									new HGlue,
									new HBox(
										textLabel = new VMessage("TextEntryLabel", qvalue->label(), Center),
										new HGlue(5, 0),
										new Frame(txt, 1)
									)
								)
							);
							enableList->Insert(textLabel, buttonValue);
							enableList->Insert(txt, buttonValue);
							txt->enable(enabled);
							textLabel->enable(enabled);
							// add current TextInput to the linked list of TextInputs
							ipd->addTextInput(txt);
						}
						break;
					default:
						break;
					}
					buttonValue <<= 1;	// next button's value
				}
				exBox->Insert(new VGlue(2, 0));
				// Insert scene containing this set of exclusive choices.
				entryBox->Insert(exBox);
				textBox->Insert(entryBox);
			}
			break;
		default:
			break;
		};
	}
	return textBox;
}

Scene *
DialogConstructor::choiceButtonBox(QueryList *list, ButtonState* state) {
	VBox* outerBox = new VBox;
	boolean notFirst = false;
	for(list->start(); list->more(); list->next()) {
		if(notFirst) outerBox->Insert(new HBorder);
		outerBox->Insert(new ButtonArray(state, (QueryChoice *) list->item()));
		notFirst = true;
	}
	outerBox->Insert(new VGlue);
	return outerBox;
}

// create set of buttons initialized with appropriate responses (Yes, No, etc.)

Scene *
DialogConstructor::buttonBox(ButtonList *list, ButtonState *s) {
	HBox* btnBox = new HBox;
	for(list->start(); list->more(); list->next()) {
		btnBox->Insert(new HGlue);
		Response resp = list->item()->response();
		Interactor *button = new ResponseButton(
			list->item()->label(),
			s,
			int(resp),
			resp == list->defaultResponse()	// true if is default
		);
		btnBox->Insert(button);
	}
	btnBox->Insert(new HGlue);
	return btnBox;
}

