// osx_dac.C

/******************************************************************************
 *
 *  MiXViews - an X window system based sound & data editor/processor
 *
 *  Copyright (c) 1993, 1994 Regents of the University of California
 *
 *  Author:     Douglas Scott
 *  Date:       December 13, 1994
 *
 *  Permission to use, copy and modify this software and its documentation
 *  for research and/or educational purposes and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation. The author reserves the right to distribute this
 *  software and its documentation.  The University of California and the author
 *  make no representations about the suitability of this software for any 
 *  purpose, and in no event shall University of California be liable for any
 *  damage, loss of data, or profits resulting from its use.
 *  It is provided "as is" without express or implied warranty.
 *
 ******************************************************************************/

#ifdef MACOSX

#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <InterViews/action.h>
#include "application.h"
#include "osx_dac.h"
#include "typeconvert.h"
#include "progressaction.h"
#include "MyString.h"

#undef DEBUG

static int getDeviceList(bool recording, AudioDeviceID **devList, int *devCount);
static int	getDeviceNames(String &devList, bool recording);
static const char *errToString(OSStatus err);

class MyTimer {
public:
	MyTimer(OSXConverter *obj) : _obj(obj) {}
	~MyTimer() {}
	void run(ProgressAction *action);
private:
	OSXConverter	*_obj;
};

void
MyTimer::run(ProgressAction *action)
{
#ifdef DEBUG
	fprintf(stderr, "MyTimer::run() starting\n");
#endif
	while(_obj->running())
	{
		float percent = _obj->_bytesDone / (float) _obj->_bytesToDo;
#ifdef DEBUG
		fprintf(stderr, "MyTimer::run() before action - percent %g\n", percent);
#endif
		if ((*action)(percent)) {
			_obj->stop();
			break;
		}
		usleep(50000);
	}
}

OSXConverter::OSXConverter()
	: _recDeviceID(kAudioDeviceUnknown), _playDeviceID(kAudioDeviceUnknown),
	  _audioBufferSize(0), _timer(new MyTimer(this)) {
	CFRunLoopSourceRef nullRunLoop = 0;
	AudioHardwareAddRunLoopSource(nullRunLoop);
}

OSXConverter::~OSXConverter() {
	stop();
	delete _timer;
}

int
OSXConverter::currentPlayLevel() const {
	return 0;
}

int
OSXConverter::currentRecordLevel() const {
	return 0;
}

boolean
OSXConverter::setPlayLevel(int volume) {
	return true;
}

boolean
OSXConverter::setRecordLevel(int volume) {
	return true;
}

boolean
OSXConverter::setSpeakerOutput(boolean toOn) {
	return true;
}

int
OSXConverter::currentInputDevice() {
	int currentDevice = 0;
	if (_recDeviceID == kAudioDeviceUnknown) {
		UInt32 size = sizeof(_recDeviceID);
		OSStatus err = AudioHardwareGetProperty(
							kAudioHardwarePropertyDefaultInputDevice,
							&size,
					  		(void *) &_recDeviceID);
		if (err != kAudioHardwareNoError || _recDeviceID == kAudioDeviceUnknown) {
			return displayError(err, "Cannot find default audio input device.");
		}
	}
	int deviceCount = 0;
	AudioDeviceID *idList = NULL;
	if (::getDeviceList(true, &idList, &deviceCount) == 0) {
		for (int n = 0; n < deviceCount; ++n) {
			if (idList[n] == _recDeviceID) {
				currentDevice = n;
				break;
			}
		}
		delete [] idList;
	}
#ifdef DEBUG
	printf("returning current input device index %d for ID %u\n", currentDevice, (unsigned)_recDeviceID);
#endif
	return currentDevice;
}

int
OSXConverter::currentOutputDevice() {
	int currentDevice = 0;
	if (_playDeviceID == kAudioDeviceUnknown) {
		UInt32 size = sizeof(_playDeviceID);
		OSStatus err = AudioHardwareGetProperty(
						kAudioHardwarePropertyDefaultOutputDevice,
						&size,
					   (void *) &_playDeviceID);
		if (err != kAudioHardwareNoError || _playDeviceID == kAudioDeviceUnknown) {
			return displayError(err, "Cannot find default audio output device.");
		}
	}
	int deviceCount = 0;
	AudioDeviceID *idList = NULL;
	if (::getDeviceList(false, &idList, &deviceCount) == 0) {
		for (int n = 0; n < deviceCount; ++n) {
			if (idList[n] == _playDeviceID) {
				currentDevice = n;
				break;
			}
		}
		delete [] idList;
	}
#ifdef DEBUG
	printf("returning current output device index %d for ID %u\n", currentDevice, (unsigned)_playDeviceID);
#endif
	return currentDevice;
}

boolean
OSXConverter::setInputDevice(int dev) {
	int deviceCount = 0;
	AudioDeviceID *idList = NULL;
	if (::getDeviceList(true, &idList, &deviceCount) == 0) {
		if (dev < deviceCount)
			_recDeviceID = idList[dev];
#ifdef DEBUG
		printf("set input device to ID %u for index %d\n", (unsigned)_recDeviceID, dev);
#endif
		delete [] idList;
		return true;
	}
	return false;
}

boolean
OSXConverter::setOutputDevice(int dev) {
	int deviceCount = 0;
	AudioDeviceID *idList = NULL;
	if (::getDeviceList(false, &idList, &deviceCount) == 0) {
		if (dev < deviceCount)
			_playDeviceID = idList[dev];
#ifdef DEBUG
		printf("set output device to ID %u for index %d\n", (unsigned)_playDeviceID, dev);
#endif
		delete [] idList;
		return true;
	}
	return false;
}

ChoiceValue
OSXConverter::inputDeviceToChoice(int iDev) {
    ChoiceValue choice = 0x1 << iDev;
    return choice;
}

ChoiceValue
OSXConverter::outputDeviceToChoice(int oDev) {
    ChoiceValue choice = 0x1 << oDev;
    return choice;
}

int
OSXConverter::inputChoiceToDevice(ChoiceValue val) {
    int dev = (int) log2(val);
    return dev;
}

int
OSXConverter::outputChoiceToDevice(ChoiceValue val) {
    int dev = (int) log2(val);
    return dev;
}

void
OSXConverter::getInputDeviceLabels(String &labels) {
    ::getDeviceNames(labels, /* recording = */ true);
}

void
OSXConverter::getOutputDeviceLabels(String &labels) {
    ::getDeviceNames(labels, /* recording = */ false);
}

boolean
OSXConverter::isPlayableFormat(DataType type) {
	return type != MuLawData;
}

int
OSXConverter::checkDataType(DataType type) {
    switch(type) {
    case FloatData:
    case DoubleData:
    case IntData:
    case ShortData:
    case SignedCharData:
    case UnsignedCharData:
		return true;
        break;
	default:
		Application::alert("Unsupported sample type");
		return false;
    }
}

int
OSXConverter::checkSampleRate(int sampleRate) {
    return true;
}

int
OSXConverter::checkChannels(int chans) {
    if (chans == 1 || chans == 2)
		return true;
	Application::alert("Unsupported channel count");
	return false;
}

int
OSXConverter::doConversion(ProgressAction* progress) {
	_bytesDone = 0;
	// Start the device.
#ifdef DEBUG
	printf("doConversion: calling AudioDeviceStart\n");
#endif
	OSStatus err = AudioDeviceStart(_playDeviceID, audioPlayProc);
	int status = (err == kAudioHardwareNoError);
	if (status == true) {
		_timer->run(progress);
	}
	else {
		displayError(err, "AudioDeviceStart [play] failed");
		stop();
	}
	return status;
}

int
OSXConverter::doRecording(ProgressAction* progress) {
	_bytesDone = 0;
	// Start the device.
#ifdef DEBUG
	printf("doRecording: calling AudioDeviceStart\n");
#endif
	OSStatus err = AudioDeviceStart(_recDeviceID, audioRecordProc);
	int status = (err == kAudioHardwareNoError);
	if (status == true) {
		_timer->run(progress);
	}
	else {
		displayError(err, "AudioDeviceStart [record] failed");
		stop();
	}
	return status;
}

int
OSXConverter::pause() {
	return true;
}

int
OSXConverter::resume() {
	return true;
}

int
OSXConverter::stop() {
	OSStatus err;
	if (recording()) {
		err = AudioDeviceStop(_recDeviceID, audioRecordProc);
		err = AudioDeviceRemoveIOProc(_recDeviceID, audioRecordProc);
	}
	else {
		err = AudioDeviceStop(_playDeviceID, audioPlayProc);
		err = AudioDeviceRemoveIOProc(_playDeviceID, audioPlayProc);
	}
	return (err == kAudioHardwareNoError) ? Super::stop() : false;
}

int
OSXConverter::doConfigure() {
	AudioDeviceID devID = willRecord() ? _recDeviceID : _playDeviceID;
	UInt32 size;
	OSStatus err;
	if (devID == kAudioDeviceUnknown) {
		size = sizeof(devID);
		err = AudioHardwareGetProperty(
						willRecord() ? 
							kAudioHardwarePropertyDefaultInputDevice :
							kAudioHardwarePropertyDefaultOutputDevice,
						&size,
					   (void *) &devID);
		if (err != kAudioHardwareNoError || devID == kAudioDeviceUnknown) {
			return displayError(err, "Cannot find default audio device.");
		}
		if (willRecord())
			_recDeviceID = devID;
		else
			_playDeviceID = devID;
	}
	// Get current input or output format	
	size = sizeof(_deviceFormat);
	err = AudioDeviceGetProperty(devID, 
								  0,
								  willRecord(),
								  kAudioDevicePropertyStreamFormat, 
								  &size, 
								  &_deviceFormat);
	if (err != kAudioHardwareNoError) {
		return displayError(err, "Can't get audio device format");
	}
#ifdef DEBUG
	printf("doConfigure: stream format:\n\tsr %g\n\tfmtID 0x%x\n\tfmtFlags 0x%x\n\tbytesperpckt %u\n\tframesperpckt %u\n\tbytesperframe %u\n\tchannelsperframe %u\n\tbitsperchannel %u\n",
			_deviceFormat.mSampleRate, _deviceFormat.mFormatID,
			 _deviceFormat.mFormatFlags, _deviceFormat.mBytesPerPacket,
			_deviceFormat.mFramesPerPacket, _deviceFormat.mBytesPerFrame,
			_deviceFormat.mChannelsPerFrame, _deviceFormat.mBitsPerChannel);
#endif
	Boolean writeable;
	// Is format writable?
	size = sizeof(writeable);
	err = AudioDeviceGetPropertyInfo(devID, 
   									0,
								    willRecord(),
								    kAudioDevicePropertyStreamFormat,
									&size,
									&writeable);
	if (err != kAudioHardwareNoError) {
		return displayError(err, "Can't get audio device property");
	}
#ifdef DEBUG
	printf("doConfigure: stream format property is %s\n",
		   writeable ? "writeable" : "not writeable");
#endif
	if (writeable)
	{
		// Set sample format, specifically, sample rate
		AudioStreamBasicDescription sFormat = _deviceFormat;
		sFormat.mSampleRate = sampleRate();
//		sFormat.mChannelsPerFrame = willRecord() ? 1 : 2;	// always open play in stereo
		size = sizeof(sFormat);
		err = AudioDeviceSetProperty(devID,
									 NULL,
									 0,
								     willRecord(),
								     kAudioDevicePropertyStreamFormat,
									 size,
									 (void *)&sFormat);
		if (err != kAudioHardwareNoError) {
			char msg1[64];
			char msg2[64];
			printf("AudioDeviceSetProperty: error %s\n", errToString(err));
			sprintf(msg1, "Can't set HW sample rate: %s", errToString(err));
		    sprintf(msg2, "Set file's sample rate to %g to %s.",
				   	_deviceFormat.mSampleRate, willRecord() ? "record" : "play");
			return error(msg1, msg2);
		}
	}
	size = sizeof(writeable);
	// Set buffer size
	err = AudioDeviceGetPropertyInfo(devID, 
   									0,
								    willRecord(),
								    kAudioDevicePropertyBufferSize, 
									&size,
									&writeable);
	if (err != kAudioHardwareNoError) {
		return displayError(err, "Can't get audio device property");
	}
	// default buffer size is 0.1 second's worth of sound
	float bufferTime = 0.1;
	const char *res = Application::getGlobalResource("OSXBufferTime");
	if (res)
		bufferTime = atof(res);
	// We cannot set the buffer size larger than the segment we wish to play.
	if (duration() < bufferTime)
		bufferTime = duration();
	// Audio buffer is always floating point
	unsigned int bufferSize = sizeof(float) *
					iround(float(channels()) * sampleRate() * bufferTime);
#ifdef DEBUG
	fprintf(stderr, "requesting buffer size %d\n", bufferSize);
#endif

	if (writeable) {
		size = sizeof(bufferSize);
		err = AudioDeviceSetProperty(devID,
									 NULL,
									 0,
								     willRecord(),
								     kAudioDevicePropertyBufferSize,
									 size,
									 (void *)&bufferSize);
		// NB: even if err != kAudioHardwareNoError, we continue and use
		//  the default buffer size.
	}
	// Get the actual buffer size.  (The device may not want to change.)
	size = sizeof(_audioBufferSize);
	err = AudioDeviceGetProperty(devID,
								 0,
								 willRecord(),
								 kAudioDevicePropertyBufferSize,
								 &size,
								 &_audioBufferSize);
	if (err != kAudioHardwareNoError) {
		return displayError(err, "Can't get audio output device buffer size");
	}
#ifdef DEBUG
	fprintf(stderr, "actual buffer size %u\n", (unsigned)_audioBufferSize);
#endif
	// Register our callback function with the HAL.
	err = AudioDeviceAddIOProc(devID,
							   willRecord() ? audioRecordProc : audioPlayProc, 
							   (void *) this);
	if (err != kAudioHardwareNoError) {
		return displayError(err, "Cannot register callback function with device.");
	}
	_bytesToDo = dataSize();
	return true;
}

OSStatus
OSXConverter::audioRecordProc(AudioDeviceID			inDevice,
							  const AudioTimeStamp	*inNow,
							  const AudioBufferList	*inInputData,
							  const AudioTimeStamp	*inInputTime,
							  AudioBufferList		*outOutputData,
							  const AudioTimeStamp	*inOutputTime,
							  void					*object)
{
	OSXConverter *converter = (OSXConverter *) object;
	unsigned inBytes = converter->_audioBufferSize;
	const DataType dataType = converter->dataType();
	const int sampSize = type_to_sampsize(dataType);
	const int inChans = converter->_deviceFormat.mChannelsPerFrame;
	const int sndChans = converter->channels();
	const int outFrames = (converter->_bytesToDo - converter->_bytesDone) / (sampSize * sndChans);
	int frames = inBytes / (inChans * sizeof(float));	// always stereo
	unsigned offset = converter->_bytesDone;

	if (outFrames < frames)
		frames = outFrames;

#ifdef DEBUG
	printf("audioRecordProc: aud buf sz: %d, bytesdone = %u bytestotal = %u\n",
			inBytes, offset, converter->_bytesToDo);
#endif

	if (frames <= 0) {
#ifdef DEBUG
		printf("audioRecordProc: calling stop()\n");
#endif
		converter->stop();
		return kAudioHardwareNoError;
	}

	char *sound = (char *)converter->pointerToData();
	char *toBuf = sound + offset;
	float *inbuf = (float *) inInputData->mBuffers[0].mData;
	double scale;
	int in;

	switch (dataType) {
	case DoubleData:
	{
		double *to = (double *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
			case 1:
				scale = 32767.0;
				for (in=0; in < frames; ++in) {
					to[in] = (double) (inbuf[in] * scale);
				}
				break;
			case 2:
				scale = 32767.0 * 0.707;
				for (in=0; in < frames; ++in) {
					to[in] = (double) (inbuf[0] + inbuf[1]) * scale;
					inbuf += 2;
				}
				break;
			}
		}
		else {
			switch (inChans) {
				case 1:
					scale = 32767.0;
					// Copy to only left channel
					for (in=0; in < frames; ++in) {
						to[0] = (double) (inbuf[in] * scale);
						to[1] = 0.0;
						to += 2;
					}
					break;
				case 2:
					scale = 32767.0;
					for (in=0; in < frames; ++in) {
						to[0] = (double)(inbuf[0] * scale);	
						to[1] = (double)(inbuf[1] * scale);	
						inbuf += 2;
						to += 2;
					}
					break;
			}
		}
	}
		break;
	case FloatData:
	{
		float *to = (float *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
			case 1:
				scale = 32767.0;
				for (in=0; in < frames; ++in) {
					to[in] = (float) (inbuf[in] * scale);
				}
				break;
			case 2:
				scale = 32767.0 * 0.707;
				for (in=0; in < frames; ++in) {
					to[in] = (float) ((inbuf[0] + inbuf[1]) * scale);
					inbuf += 2;
				}
				break;
			}
		}
		else {
			switch (inChans) {
				case 1:
					scale = 32767.0;
					// Copy to only left channel
					for (in=0; in < frames; ++in) {
						to[0] = (float) (inbuf[in] * scale);
						to[1] = 0.0f;
						to += 2;
					}
					break;
				case 2:
					scale = 32767.0;
					for (in=0; in < frames; ++in) {
						to[0] = (float)(inbuf[0] * scale);	
						to[1] = (float)(inbuf[1] * scale);	
						inbuf += 2;
						to += 2;
					}
					break;
			}
		}
	}
		break;
	case IntData:
	{
		int *to = (int *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
			case 1:
				scale = MAXINT;
				for (in=0; in < frames; ++in) {
					to[in] = (int) (inbuf[in] * scale);
				}
				break;
			case 2:
				scale = MAXINT * 0.707;
				for (in=0; in < frames; ++in) {
					to[in] = (int) ((inbuf[0] + inbuf[1]) * scale);
					inbuf += 2;
				}
				break;
			}
		}
		else {
			switch (inChans) {
				case 1:
					scale = MAXINT;
					// Copy to only left channel
					for (in=0; in < frames; ++in) {
						to[0] = (int) (inbuf[in] * scale);
						to[1] = 0;
						to += 2;
					}
					break;
				case 2:
					scale = MAXINT;
					for (in=0; in < frames; ++in) {
						to[0] = (int)(inbuf[0] * scale);	
						to[1] = (int)(inbuf[1] * scale);	
						inbuf += 2;
						to += 2;
					}
					break;
			}
		}
	}
		break;
	case ShortData:
	{
		short *to = (short *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
				case 1:
					scale = 32767.0;
					for (in=0; in < frames; ++in) {
						to[in] = (short) (inbuf[in] * scale);
					}
					break;
				case 2:
					scale = 32767.0 * 0.707;
					for (in=0; in < frames; ++in) {
						to[in] = (short) ((inbuf[0] + inbuf[1]) * scale);
						inbuf += 2;
					}
					break;
			}
		}
		else {
			switch (inChans) {
				case 1:
					scale = 32767.0;
					// Copy to only left channel
					for (in=0; in < frames; ++in) {
						to[0] = (short) (inbuf[in] * scale);
						to[1] = 0;
						to += 2;
					}
					break;
				case 2:
					scale = 32767.0;
					for (in=0; in < frames; ++in) {
						to[0] = (short)(inbuf[0] * scale);	
						to[1] = (short)(inbuf[1] * scale);	
						inbuf += 2;
						to += 2;
					}
					break;
			}
		}
	}
		break;
	case SignedCharData:
	{
		signed char *to = (signed char *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
			case 1:
				scale = 127.0;
				for (in=0; in < frames; ++in) {
					to[in] = (signed char) (inbuf[in] * scale);
				}
				break;
			case 2:
				scale = 127.0 * 0.707;
				for (in=0; in < frames; ++in) {
					to[in] = (signed char) ((inbuf[0] + inbuf[1]) * scale);
					inbuf += 2;
				}
				break;
			}
		}
		else {
			scale = 127.0;
			for (in=0; in < frames; ++in) {
				to[0] = (signed char)(inbuf[0] * scale);	
				to[1] = (signed char)(inbuf[1] * scale);	
				inbuf += 2;
				to += 2;
			}
		}
	}
		break;
	case UnsignedCharData:
	{
		signed char *to = (signed char *) toBuf;
		if (sndChans == 1) {
			switch (inChans) {
			case 1:
				scale = 127.0;
				for (in=0; in < frames; ++in) {
					to[in] = (unsigned char) (inbuf[in] * scale) + 0x80;
				}
				break;
			case 2:
				scale = 127.0 * 0.707;
				for (in=0; in < frames; ++in) {
					to[in] = (unsigned char) ((inbuf[0] + inbuf[1]) * scale) + 0x80;
					inbuf += 2;
				}
			}
		}
		else {
			scale = 127.0;
			for (in=0; in < frames; ++in) {
				to[0] = (unsigned char)(inbuf[0] * scale) + 0x80;
				to[1] = (unsigned char)(inbuf[1] * scale) + 0x80;	
				inbuf += 2;
				to += 2;
			}
		}
	}
		break;
	default:
		exit(1);
	}

	unsigned outBytes = frames * sndChans * sampSize;
//	inInputData->mBuffers[0].mDataByteSize = outBytes;
	converter->_bytesDone += outBytes;
	return kAudioHardwareNoError;
}

OSStatus
OSXConverter::audioPlayProc(AudioDeviceID			inDevice,
						  const AudioTimeStamp	*inNow,
						  const AudioBufferList	*inInputData,
						  const AudioTimeStamp	*inInputTime,
						  AudioBufferList		*outOutputData,
						  const AudioTimeStamp	*inOutputTime,
						  void					*object)
{
	OSXConverter *converter = (OSXConverter *) object;
	unsigned outBytes = converter->_audioBufferSize;
	const DataType dataType = converter->dataType();
	const int sampSize = type_to_sampsize(dataType);
	const int sndChans = converter->channels();
	const int inFrames = (converter->_bytesToDo - converter->_bytesDone) / (sampSize * sndChans);
	int frames = outBytes / (2 * sizeof(float));	// always stereo
	unsigned offset = converter->_bytesDone;

	if (inFrames < frames)
		frames = inFrames;

#ifdef DEBUG
	printf("audioPlayProc: aud buf sz: %d, bytesdone = %u bytestotal = %u\n",
			outBytes, offset, converter->_bytesToDo);
#endif

	if (frames <= 0) {
#ifdef DEBUG
		printf("audioPlayProc: calling stop()\n");
#endif
		converter->stop();
		return kAudioHardwareNoError;
	}

	char *sound = (char *)converter->pointerToData();
	char *frombuf = sound + offset;
	float *outbuf = (float *) outOutputData->mBuffers[0].mData;
	double scale;
	int in;

	switch (dataType) {
	case DoubleData:
	{
		scale = 1.0/converter->peakAmplitude();
		double *from = (double *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)(from[in] * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)(from[0] * scale);	
				outbuf[1] = (float)(from[1] * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	case FloatData:
	{
		scale = 1.0/converter->peakAmplitude();
		float *from = (float *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)(from[in] * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)(from[0] * scale);	
				outbuf[1] = (float)(from[1] * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	case IntData:
	{
		scale = 1.0/MAXINT;
		int *from = (int *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)(from[in] * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)(from[0] * scale);	
				outbuf[1] = (float)(from[1] * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	case ShortData:
	{
		scale = 1.0/MAXSHORT;
		short *from = (short *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)(from[in] * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)(from[0] * scale);	
				outbuf[1] = (float)(from[1] * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	case SignedCharData:
	{
		scale = 1.0/128;
		signed char *from = (signed char *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)(from[in] * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)(from[0] * scale);	
				outbuf[1] = (float)(from[1] * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	case UnsignedCharData:
	{
		scale = 1.0/128;
		signed char *from = (signed char *) frombuf;
		if (sndChans == 1)
			for (in=0; in < frames; ++in) {
				outbuf[1] = outbuf[0] = (float)((from[in]+0x80) * scale);	 
				outbuf += 2;
			}
		else
			for (in=0; in < frames; ++in) {
				outbuf[0] = (float)((from[0]+0x80) * scale);	
				outbuf[1] = (float)((from[1]+0x80) * scale);	
				from += 2;
				outbuf += 2;
			}
	}
		break;
	default:
		exit(1);
	}

	unsigned inBytes = frames * sndChans * sampSize;
	outOutputData->mBuffers[0].mDataByteSize = frames * 2 * sizeof(float);
	converter->_bytesDone += inBytes;
	return kAudioHardwareNoError;
}

int
OSXConverter::displayError(int err, const char *msg)
{
	return error(msg, ::errToString(err));
}

int 
OSXConverter::error(const char* msg1, const char* msg2)
{
	errno = 0;
	return RealConverter::error(msg1, msg2);
}

static int
getDeviceList(bool recording, AudioDeviceID **devList, int *devCount)
{
   UInt32 size;
   OSStatus err = AudioHardwareGetPropertyInfo(
                        kAudioHardwarePropertyDevices,
                        &size, NULL);
   if (err != kAudioHardwareNoError) {
      fprintf(stderr, "Can't get hardware device list property info.\n");
      return err;
   }
   int totalDevCount = size / sizeof(AudioDeviceID);
   AudioDeviceID *deviceList = new AudioDeviceID[totalDevCount];
   err = AudioHardwareGetProperty(
                        kAudioHardwarePropertyDevices,
                        &size, deviceList);
   if (err != kAudioHardwareNoError) {
      fprintf(stderr, "Can't get hardware device list.\n");
      return err;
   }
   
#ifdef DEBUG
	printf("Full Device ID list: ");
	for (int z = 0; z < totalDevCount; ++z) printf("%u  ", deviceList[z]);
	printf("\n");
#endif
	for (int n = 0; n < totalDevCount; ++n) {
   		Boolean writeable = 0;
		AudioBufferList *list = NULL;
   		size = 0;	// reset
		// Get size of AudioBufferList, in bytes.
		err = AudioDeviceGetPropertyInfo(deviceList[n],
                    		   0, recording,
                    		   kAudioDevicePropertyStreamConfiguration,
                    		   &size,
                    		   &writeable);
		if (!err && size > 0) {
			list = new AudioBufferList [size / sizeof(AudioBufferList)];
			list->mBuffers[0].mNumberChannels = 0;
			err = AudioDeviceGetProperty(deviceList[n],
										 0, recording,
										 kAudioDevicePropertyStreamConfiguration,
										 &size,
										 list);
		}
		// Prune devices which are not for the direction we are currently requesting.
		if (err != kAudioHardwareNoError || !list || list->mBuffers[0].mNumberChannels == 0) {
#ifdef DEBUG
			printf("Removing device ID %u from %s list\n", deviceList[n], recording ? "record" : "playback");
#endif
        	for (int i = n; i < totalDevCount - 1; ++i) {
        		deviceList[i] = deviceList[i+1];
				deviceList[i+1] = 0;
			}
			--n;
			--totalDevCount;
        }
		delete [] list;
    }
#ifdef DEBUG
	printf("Device ID list for %s: ", recording ? "record" : "playback");
	for (int x = 0; x < totalDevCount; ++x) printf("%u  ", deviceList[x]);
	printf("\n");
#endif
	*devList = deviceList;
	*devCount = totalDevCount;
	return 0;
}

static int	getDeviceNames(String &devList, bool recording)
{
	int deviceCount = 0;
	int nameCount = 0;
	AudioDeviceID *idList = NULL;
	devList += "|";
	::getDeviceList(recording, &idList, &deviceCount);
	for (int dev = 0; dev < deviceCount; ++dev) {
		UInt32 size;
		char *name = NULL, descriptor[32];
		OSStatus err = AudioDeviceGetPropertyInfo(idList[dev],
                                        		 0,
                                        		 recording,
                                        		 kAudioDevicePropertyDeviceName,
                                        		 &size, NULL);
		if (err == kAudioHardwareNoError) {
			name = new char[size + 1];
			err = AudioDeviceGetProperty(idList[dev],
                            			0,
                            			recording,
                            			kAudioDevicePropertyDeviceName,
                            			&size, name);
			if (err == kAudioHardwareNoError) {
				snprintf(descriptor, 31, "%s", name);
				descriptor[31] = '\0';
			}
			else {
				sprintf(name, "(No name) %d", dev);
				sprintf(descriptor, "default");
			}
		}
		else {
			name = new char[12];
			sprintf(name, "(No name) %d", dev);
			sprintf(descriptor, "default");
		}

		
		String str;
		char tempBuffer[64];
   		Boolean writeable = 0;
		// Get size of AudioBufferList, in bytes.
		err = AudioDeviceGetPropertyInfo(idList[dev],
                    		   0, recording,
                    		   kAudioDevicePropertyStreamConfiguration,
                    		   &size,
                    		   &writeable);
		if (err == kAudioHardwareNoError) {
			UInt32 currentChan = 1;
			// Fill list with description of buffers.
			AudioBufferList *list = new AudioBufferList [size / sizeof(AudioBufferList)];
			err = AudioDeviceGetProperty(idList[dev],
                    			   0, recording,
                    			   kAudioDevicePropertyStreamConfiguration,
                    			   &size,
                    			   list);
			if (err == kAudioHardwareNoError) {
				// Catch special case where device consists of multiple mono streams.
				if (list->mBuffers[0].mNumberChannels == 1) {
					sprintf(tempBuffer, "%s: (%lu channels)",
								name, list->mNumberBuffers);
					str = tempBuffer;
					devList += str;
					devList += "|";
					++nameCount;
#ifdef DEBUG
					fprintf(stderr, "GetDeviceNames: created item \"%s\"\n", str.chars());
#endif
				}
				// Otherwise we have a set of independent, interleaved streams.
				else for (unsigned stream = 0; stream < list->mNumberBuffers; ++stream)
				{
					String subdev = "";
					// See if we can get a subdevice name.
					err = AudioDeviceGetPropertyInfo(idList[dev],
                                             currentChan,
                                             recording,
                                             kAudioDevicePropertyChannelName,
                                             &size, NULL);
					if (err == kAudioHardwareNoError) {
						char *chanName = new char[size + 1];
						err = AudioDeviceGetProperty(idList[dev],
                            						currentChan,
                            						recording,
                            						kAudioDevicePropertyChannelName,
                            						&size, chanName);
						if (err == kAudioHardwareNoError) {
							subdev = chanName;
						}
						delete [] chanName;
					}
					if (subdev != "")
						sprintf(tempBuffer, "%s: [%s] (%lu channels)",
									name, (const char *) subdev,
									list->mBuffers[stream].mNumberChannels);
					else
						sprintf(tempBuffer, "%s: (%lu channels)",
									name, list->mBuffers[stream].mNumberChannels);
					str = tempBuffer;
					devList += str;
					devList += "|";
					++nameCount;
#ifdef DEBUG
					fprintf(stderr, "GetDeviceNames: created item \"%s\"\n", str.chars());
#endif
					currentChan += list->mBuffers[stream].mNumberChannels;
				}
			}
			else {
				fprintf(stderr, "Can't get device (%lu) stream config: %s\n",
					   idList[dev], ::errToString(err));
				str = name;
				devList += str;
				devList += "|";
				++nameCount;
			}
			delete [] list;
		}
		else {
			fprintf(stderr, "Can't get device (%lu) stream config property info: %s\n",
        		    idList[dev], ::errToString(err));
			str = name;
			devList += str;
			devList += "|";
			++nameCount;
		}
		delete [] name;
	}
	return nameCount;
}

// Static utilities

static const char *
errToString(OSStatus err)
{
	const char *errstring;
	switch (err) {
	case kAudioHardwareNoError:
		errstring = "No error.";
		break;
	case kAudioHardwareNotRunningError:
		errstring = "Hardware not running.";
		break;
	case kAudioHardwareUnspecifiedError:
		errstring = "Unspecified error.";
		break;
	case kAudioHardwareUnknownPropertyError:
		errstring = "Unknown hardware property.";
		break;
	case kAudioDeviceUnsupportedFormatError:
		errstring = "Unsupported audio format.";
		break;
	case kAudioHardwareBadPropertySizeError:
		errstring = "Bad hardware propery size.";
		break;
	case kAudioHardwareIllegalOperationError:
		errstring = "Illegal operation.";
		break;
	default:
		errstring = "Unknown error.";
	}
	return errstring;
}

#endif	/* MACOSX */
