// request.h

/******************************************************************************
 *
 *  MiXViews - an X window system based sound & data editor/processor
 *
 *  Copyright (c) 1993, 1994 Regents of the University of California
 *
 *  Author:     Douglas Scott
 *  Date:       December 13, 1994
 *
 *  Permission to use, copy and modify this software and its documentation
 *  for research and/or educational purposes and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation. The author reserves the right to distribute this
 *  software and its documentation.  The University of California and the author
 *  make no representations about the suitability of this software for any 
 *  purpose, and in no event shall University of California be liable for any
 *  damage, loss of data, or profits resulting from its use.
 *  It is provided "as is" without express or implied warranty.
 *
 ******************************************************************************/


// a Request is sent from a Requestor to a Controller, giving it a set of
// QueryItems, which contain variables to assign values to, labels for querying
// for those variables, and a pointer to a function to check the entered chars.
// The Controller then passes the Request to the DialogConstructor, where the
// QueryItem instances are passed to the GUI components.  After the parameters
// are set by the user, the Request hands those values back to the Requester.
// Many Requests are merely to display a message for the user, or ask for
// confirmation of some operation, or allow a simple choice.  These contain
// "read-only" QueryItems which contain text strings to display.
// See query.h, requester.h, and dialog_ctor.h.

#ifndef REQUEST_H
#ifdef __GNUG__
#pragma interface
#endif
#define REQUEST_H

#include "localdefs.h"
#include <InterViews/resource.h>
#include "range.h"

class String;

enum RequestType { AlertType=0, ConfirmerType, ChoiceType, InputType };

struct QueryInfo;
struct QueryLabelInfo;
struct QueryButtonInfo;
class DialogConstructor;
class QueryFile;
class QueryList;
class ButtonList;
class QueryValue;
class QueryItem;

class Request : public Resource {
	friend class DialogConstructor;
public:
	Request(const QueryInfo *);
	Request(const char *label);
	virtual ~Request();
	
	boolean hasLabels() { return labelList != nil; }
	boolean hasBrowser() { return browserQuery != nil; }
	boolean hasValues() { return valueList != nil; }
	boolean hasChoices() { return choiceList != nil; }
	boolean useBell() { return bell; }
	virtual boolean print(FILE *);
	virtual RequestType type() { return InputType; }	// default
	void appendLabel(const char *);
	void appendLabel(const char *, const char *);
	void addFileBrowser(String* path, const char* suffixes=nil);
	// Add single value queries of various types.
	void appendValue(QueryItem *);
	void appendValue(const char *, String* value);
	void appendValue(const char *, TimeVal* time, const Range& =NonNegativeNumbers);
	void appendValue(const char *, int *, const Range& =AllIntegers,
	                 boolean show=false);
	void appendValue(const char *, unsigned short *,
			 const Range& =NonNegativeShorts, boolean show=false);
	void appendValue(const char *, float *, const Range& =AllNumbers,
	                 boolean show=false);
	void appendValue(const char *, double *, const Range& =AllNumbers,
	                 boolean show=false);
	// Add a set of value queries, mutually exclusive, with selection buttons.
	void appendChoiceValues(const char *label, QueryValue *, QueryValue *, ChoiceValue *selected);
	void appendChoiceValues(const char *label, QueryValue *, QueryValue *, QueryValue *, ChoiceValue *selected);
	// Add set of selection buttons, either exclusive or not.
	void appendChoice(const char *, const char *lbls, ChoiceValue* states,
		boolean excl=true);
	void appendButton(const char *, Response);
	void setBell(boolean b) { bell = b; }
protected:
	Request() { init(); }	// used by derived classes
	void newLabelList();
	void newButtonList();
	virtual void createLabelList(QueryLabelInfo *);
	virtual void createButtonList(QueryButtonInfo *, Response r=Yes);
protected:
	QueryList* labelList;
	QueryFile* browserQuery;
	QueryList* valueList;
	QueryList* choiceList;
	ButtonList* buttonList;
private:
	void init();
private:
	boolean bell;							// flag for bell ring
};

// subclass for requests which open browser for file selection

class FileRequest : public Request {
public:
	FileRequest(const char* title, String* dir, const char* suffix=nil);
};

// intermediate, abstract base class for following classes

class MessageRequest : public Request {
protected:
	MessageRequest(const char* m1, const char* m2, const char* m3);
};

// AlertRequest only asks for a message display and a single "ok" button

class AlertRequest : public MessageRequest {
	typedef MessageRequest Super;
	friend class DialogConstructor;
public:
	AlertRequest(const char* m1, const char* m2=nil, const char* m3=nil,
		const char* ="  ok  ");
	virtual RequestType type() { return AlertType; }
private:
	void init(const char*);
};

// ConfirmRequest asks for a message display and a confirm-cancel button pair

class ConfirmRequest : public MessageRequest {
	typedef MessageRequest Super;
	friend class DialogConstructor;
public:
	ConfirmRequest(const char *m1, const char *m2=nil,
		const char *m3=nil, Response r=Yes,
		const char* ="confirm", const char* ="cancel");
	virtual RequestType type() { return ConfirmerType; }
private:
	void init(Response, const char*, const char*);
};

// ChoiceRequest asks for a message display and yes, no, and cancel buttons

class ChoiceRequest : public MessageRequest {
	typedef MessageRequest Super;
	friend class DialogConstructor;
public:
	ChoiceRequest(const char *m1, const char *m2=nil, const char *m3=nil,
		Response r=Yes, const char* =" yes ", const char* ="  no  ",
		const char* ="cancel");
	virtual RequestType type() { return ChoiceType; }
private:
	void init(Response, const char*, const char*, const char*);
};

extern QueryButtonInfo defaultInputButtonInfo[];

#endif

